import functools
from typing import List, Sequence

import distrax
import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from flax.linen.initializers import constant, orthogonal

from kinetix.environment.spaces import ActionType
from kinetix.models.action_spaces import HybridActionDistribution, MultiDiscreteActionDistribution


class ScannedRNN(nn.Module):
    @functools.partial(
        nn.scan,
        variable_broadcast="params",
        in_axes=0,
        out_axes=0,
        split_rngs={"params": False},
    )
    @nn.compact
    def __call__(self, carry, x):
        """Applies the module."""
        rnn_state = carry
        ins, resets = x
        rnn_state = jnp.where(
            resets[:, np.newaxis],
            self.initialize_carry(ins.shape[0], 256),
            rnn_state,
        )
        new_rnn_state, y = nn.GRUCell(features=256)(rnn_state, ins)
        return new_rnn_state, y

    @staticmethod
    def initialize_carry(batch_size, hidden_size=256):
        # Use a dummy key since the default state init fn is just zeros.
        cell = nn.GRUCell(features=256)
        return cell.initialize_carry(jax.random.PRNGKey(0), (batch_size, hidden_size))


class GeneralActorCriticRNN(nn.Module):
    action_dim: Sequence[int]
    fc_layer_depth: int
    fc_layer_width: int
    action_type: ActionType
    hybrid_action_continuous_dim: int
    multi_discrete_number_of_dims_per_distribution: List[int]
    add_generator_embedding: bool = False
    generator_embedding_number_of_timesteps: int = 10
    recurrent: bool = False

    # Given an embedding, return the action/values, since this is shared across all models.
    @nn.compact
    def __call__(self, hidden, obs, embedding, dones, activation):

        if self.add_generator_embedding:
            raise NotImplementedError()

        if self.recurrent:
            rnn_in = (embedding, dones)
            hidden, embedding = ScannedRNN()(hidden, rnn_in)

        actor_mean = embedding
        critic = embedding
        actor_mean_last = embedding
        for _ in range(self.fc_layer_depth):
            actor_mean = nn.Dense(
                self.fc_layer_width,
                kernel_init=orthogonal(np.sqrt(2)),
                bias_init=constant(0.0),
            )(actor_mean)
            actor_mean = activation(actor_mean)

            critic = nn.Dense(
                self.fc_layer_width,
                kernel_init=orthogonal(np.sqrt(2)),
                bias_init=constant(0.0),
            )(critic)
            critic = activation(critic)

        actor_mean_last = actor_mean
        actor_mean = nn.Dense(self.action_dim, kernel_init=orthogonal(0.01), bias_init=constant(0.0))(actor_mean)
        if self.action_type == ActionType.DISCRETE:
            pi = distrax.Categorical(logits=actor_mean)
        elif self.action_type == ActionType.CONTINUOUS:
            actor_logtstd = self.param("log_std", nn.initializers.zeros, (self.action_dim,))
            pi = distrax.MultivariateNormalDiag(actor_mean, jnp.exp(actor_logtstd))
        elif self.action_type == ActionType.MULTI_DISCRETE:
            pi = MultiDiscreteActionDistribution(actor_mean, self.multi_discrete_number_of_dims_per_distribution)
        else:
            actor_mean_continuous = nn.Dense(
                self.hybrid_action_continuous_dim, kernel_init=orthogonal(0.01), bias_init=constant(0.0)
            )(actor_mean_last)
            actor_mean_sigma = jnp.exp(
                nn.Dense(self.hybrid_action_continuous_dim, kernel_init=orthogonal(0.01), bias_init=constant(0.0))(
                    actor_mean_last
                )
            )
            pi = HybridActionDistribution(actor_mean, actor_mean_continuous, actor_mean_sigma)

        critic = nn.Dense(1, kernel_init=orthogonal(1.0), bias_init=constant(0.0))(critic)
        return hidden, pi, jnp.squeeze(critic, axis=-1)


class ActorCriticPixelsRNN(nn.Module):

    action_dim: Sequence[int]
    fc_layer_depth: int
    fc_layer_width: int
    action_mode: str
    hybrid_action_continuous_dim: int
    multi_discrete_number_of_dims_per_distribution: List[int]
    activation: str
    add_generator_embedding: bool = False
    generator_embedding_number_of_timesteps: int = 10
    recurrent: bool = True

    @nn.compact
    def __call__(self, hidden, x, **kwargs):
        if self.activation == "relu":
            activation = nn.relu
        else:
            activation = nn.tanh
        og_obs, dones = x

        if self.add_generator_embedding:
            obs = og_obs.obs
        else:
            obs = og_obs

        image = obs.image
        global_info = obs.global_info

        x = nn.Conv(features=16, kernel_size=(8, 8), strides=(4, 4))(image)
        x = nn.relu(x)
        x = nn.Conv(features=32, kernel_size=(4, 4), strides=(2, 2))(x)
        x = nn.relu(x)
        embedding = x.reshape(x.shape[0], x.shape[1], -1)

        embedding = jnp.concatenate([embedding, global_info], axis=-1)

        return GeneralActorCriticRNN(
            action_dim=self.action_dim,
            fc_layer_depth=self.fc_layer_depth,
            fc_layer_width=self.fc_layer_width,
            action_type=self.action_mode,
            hybrid_action_continuous_dim=self.hybrid_action_continuous_dim,
            multi_discrete_number_of_dims_per_distribution=self.multi_discrete_number_of_dims_per_distribution,
            add_generator_embedding=self.add_generator_embedding,
            generator_embedding_number_of_timesteps=self.generator_embedding_number_of_timesteps,
            recurrent=self.recurrent,
        )(hidden, og_obs, embedding, dones, activation)

    @staticmethod
    def initialize_carry(batch_size, hidden_size=256):
        return ScannedRNN.initialize_carry(batch_size, hidden_size)


class ActorCriticSymbolicRNN(nn.Module):
    action_dim: Sequence[int]
    fc_layer_width: int
    action_mode: str
    hybrid_action_continuous_dim: int
    multi_discrete_number_of_dims_per_distribution: List[int]
    fc_layer_depth: int
    activation: str
    add_generator_embedding: bool = False
    generator_embedding_number_of_timesteps: int = 10
    recurrent: bool = True

    @nn.compact
    def __call__(self, hidden, x):
        if self.activation == "relu":
            activation = nn.relu
        else:
            activation = nn.tanh

        og_obs, dones = x
        if self.add_generator_embedding:
            obs = og_obs.obs
        else:
            obs = og_obs

        embedding = nn.Dense(
            self.fc_layer_width,
            kernel_init=orthogonal(np.sqrt(2)),
            bias_init=constant(0.0),
        )(obs)
        embedding = nn.relu(embedding)

        return GeneralActorCriticRNN(
            action_dim=self.action_dim,
            fc_layer_depth=self.fc_layer_depth,
            fc_layer_width=self.fc_layer_width,
            action_type=self.action_mode,
            hybrid_action_continuous_dim=self.hybrid_action_continuous_dim,
            multi_discrete_number_of_dims_per_distribution=self.multi_discrete_number_of_dims_per_distribution,
            add_generator_embedding=self.add_generator_embedding,
            generator_embedding_number_of_timesteps=self.generator_embedding_number_of_timesteps,
            recurrent=self.recurrent,
        )(hidden, og_obs, embedding, dones, activation)

    @staticmethod
    def initialize_carry(batch_size, hidden_size=256):
        return ScannedRNN.initialize_carry(batch_size, hidden_size)


class MultiHeadDense(nn.Module):
    num_heads: int  # Number of heads
    out_dim: int  # Output dimension for each head
    kernel_init: nn.initializers.Initializer
    bias_init: nn.initializers.Initializer

    @nn.compact
    def __call__(self, x):
        # x has shape (...., num_features, feature_dim)
        num_features, feature_dim = x.shape[-2:]

        # Initialize a Dense layer for each head, stacked as (num_heads, feature_dim, out_dim)
        dense_kernels = self.param("dense_kernels", self.kernel_init, (self.num_heads, feature_dim, self.out_dim))
        dense_biases = self.param("dense_biases", self.bias_init, (self.num_heads, self.out_dim))

        # Apply the dense layer to each head by broadcasting and matrix multiplying
        x_expanded = jnp.expand_dims(x, axis=-2)  # Shape: (..., num_features, 1, feature_dim)
        output = jnp.einsum("...fhd,hdo->...fho", x_expanded, dense_kernels) + dense_biases
        output = nn.relu(output)  # Shape: (..., num_features, num_heads, out_dim)
        output = output.sum(axis=-3)  # Shape=(..., num_heads, out_dim)

        output = output.reshape((*output.shape[:-2], self.num_heads * self.out_dim))  # Shape=(..., num_heads * out_dim)
        return output


class ActorCriticPermutationInvariantSymbolicRNN(nn.Module):
    action_dim: Sequence[int]
    symbolic_embedding_dim: int
    fc_layer_width: int
    action_mode: str
    multi_discrete_number_of_dims_per_distribution: List[int]
    hybrid_action_continuous_dim: int
    fc_layer_depth: int
    activation: str
    recurrent: bool
    add_generator_embedding: bool = False
    include_actions_and_rewards: bool = False
    permutation_invariant: bool = True
    num_heads: int = None
    preprocess_separately: bool = False
    encoder_size: int = 64

    @nn.compact
    def __call__(self, hidden, x):
        # print(f"ActorCriticSymbolicRNN\n")
        if self.activation == "relu":
            activation_fn = nn.relu
        elif self.activation == "tanh":
            activation_fn = nn.tanh
        else:
            raise ValueError(f"Unknown activation function: {self.activation}")

        og_obs, dones = x
        if self.add_generator_embedding:
            obs = og_obs.obs
        else:
            obs = og_obs

        if self.permutation_invariant:
            assert (
                self.symbolic_embedding_dim % self.num_heads == 0
            ), f"{self.symbolic_embedding_dim=} must be divisible by {self.num_heads=}"

            if self.preprocess_separately:

                def _single_encoder(features, entity_id):
                    num_to_remove = 4
                    embedding = activation_fn(
                        nn.Dense(
                            self.encoder_size - num_to_remove,
                            kernel_init=orthogonal(np.sqrt(2)),
                            bias_init=constant(0.0),
                        )(features)
                    )
                    id_1h = jnp.zeros((*embedding.shape[:3], 4)).at[:, :, :, entity_id].set(1)
                    return jnp.concatenate([embedding, id_1h], axis=-1)

                circle_encodings = _single_encoder(obs.circles, 0)
                polygon_encodings = _single_encoder(obs.polygons, 1)
                joint_encodings = _single_encoder(obs.joints, 2)
                thruster_encodings = _single_encoder(obs.thrusters, 3)

                all_encodings = jnp.concatenate(
                    [polygon_encodings, circle_encodings, joint_encodings, thruster_encodings], axis=2
                )
                all_mask = jnp.concatenate(
                    [obs.polygon_mask, obs.circle_mask, obs.joint_mask, obs.thruster_mask], axis=2
                )

                def mask(features, mask):
                    return jnp.where(mask[:, None], features, jnp.zeros_like(features))

                obs = jax.vmap(jax.vmap(mask))(all_encodings, all_mask)

            dim_per_head = self.symbolic_embedding_dim // self.num_heads
            obs_embedding = MultiHeadDense(
                num_heads=self.num_heads,
                out_dim=dim_per_head,
                kernel_init=orthogonal(np.sqrt(2)),
                bias_init=constant(0.0),
            )(obs)
            embedding = obs_embedding
        embedding = activation_fn(embedding)

        return GeneralActorCriticRNN(
            action_dim=self.action_dim,
            fc_layer_depth=self.fc_layer_depth,
            fc_layer_width=self.fc_layer_width,
            action_type=self.action_mode,
            hybrid_action_continuous_dim=self.hybrid_action_continuous_dim,
            multi_discrete_number_of_dims_per_distribution=self.multi_discrete_number_of_dims_per_distribution,
            recurrent=self.recurrent,
        )(hidden, og_obs, embedding, dones, activation_fn)

    @staticmethod
    def initialize_carry(batch_size, hidden_dim):
        return ScannedRNN.initialize_carry(batch_size, hidden_dim)



import dataclasses
import jax.numpy as jnp
from flax import struct
from jax2d.sim_state import SimParams, SimState, StaticSimParams


@struct.dataclass
class EnvState(SimState):
    thruster_bindings: jnp.ndarray
    motor_bindings: jnp.ndarray
    motor_auto: jnp.ndarray

    polygon_shape_roles: jnp.ndarray
    circle_shape_roles: jnp.ndarray

    polygon_highlighted: jnp.ndarray
    circle_highlighted: jnp.ndarray

    polygon_densities: jnp.ndarray
    circle_densities: jnp.ndarray

    timestep: int = 0

    last_distance: float = -1.0


@struct.dataclass
class EnvParams(SimParams):
    max_timesteps: int = 256
    pixels_per_unit: int = 100
    dense_reward_scale: float = 0.1
    num_shape_roles: int = 4

    def __hash__(self):
        field_values = {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}
        return hash(tuple(sorted(field_values.items())))

    def __eq__(self, value):
        return super().__eq__(value)


@struct.dataclass
class StaticEnvParams(StaticSimParams):
    screen_dim: tuple[int, int] = (500, 500)
    downscale: int = 4

    frame_skip: int = 1
    max_shape_size: int = 2

    num_motor_bindings: int = 4
    num_thruster_bindings: int = 2

    def __hash__(self):
        field_values = {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}
        return hash(tuple(sorted(field_values.items())))

    def __eq__(self, value):
        return super().__eq__(value)



from enum import Enum
from functools import partial
from typing import Callable

import chex
import jax
import jax.numpy as jnp
from jax2d.engine import PhysicsEngine

from kinetix.environment.env_state import EnvParams, EnvState, StaticEnvParams
from kinetix.environment.ued.distributions import create_vmapped_filtered_distribution, sample_kinetix_level
from kinetix.environment.ued.mutators import (
    make_mutate_change_shape_rotation,
    make_mutate_change_shape_size,
    mutate_add_connected_shape,
    mutate_add_connected_shape_proper,
    mutate_add_shape,
    mutate_add_thruster,
    mutate_change_gravity,
    mutate_change_shape_location,
    mutate_remove_joint,
    mutate_remove_shape,
    mutate_remove_thruster,
    mutate_swap_role,
    mutate_toggle_fixture,
)
from kinetix.environment.ued.ued_state import UEDParams
from kinetix.environment.utils import create_empty_env
from kinetix.util.saving import load_evaluation_levels


def make_mutate_env(static_env_params: StaticEnvParams, env_params: EnvParams, ued_params: UEDParams):
    mutate_size = make_mutate_change_shape_size(env_params, static_env_params)
    mutate_rot = make_mutate_change_shape_rotation(env_params, static_env_params)

    def mutate_level(rng, level: EnvState, n=1):
        def inner(carry: tuple[chex.PRNGKey, EnvState], _):
            rng, level = carry
            rng, _rng, _rng2 = jax.random.split(rng, 3)

            any_rects_left = jnp.logical_not(level.polygon.active).sum() > 0
            any_circles_left = jnp.logical_not(level.circle.active).sum() > 0
            any_joints_left = jnp.logical_not(level.joint.active).sum() > 0
            any_thrust_left = jnp.logical_not(level.thruster.active).sum() > 0
            has_any_thursters = level.thruster.active.sum() > 0

            can_do_add_shape = any_rects_left | any_circles_left
            can_do_add_joint = can_do_add_shape & any_joints_left

            all_mutations = [
                mutate_add_shape,
                mutate_add_connected_shape_proper,
                mutate_remove_joint,
                mutate_remove_shape,
                mutate_swap_role,
                mutate_add_thruster,
                mutate_remove_thruster,
                mutate_toggle_fixture,
                mutate_size,
                mutate_change_shape_location,
                mutate_rot,
            ]

            def mypartial(f):
                def inner(rng, level):
                    return f(rng, level, env_params, static_env_params, ued_params)

                return inner

            probs = jnp.array(
                [
                    can_do_add_shape * 1.0,
                    can_do_add_joint * 1.0,
                    0.0,
                    0.0,
                    1.0,
                    any_thrust_left * 1.0,
                    has_any_thursters * 1.0,
                    0.1,
                    1.0,
                    1.0,
                    1.0,
                ]
            )

            all_mutations = [mypartial(i) for i in all_mutations]
            index = jax.random.choice(_rng, jnp.arange(len(all_mutations)), (), p=probs)
            level = jax.lax.switch(index, all_mutations, _rng2, level)

            return (rng, level), None

        (_, level), _ = jax.lax.scan(inner, (rng, level), None, length=n)
        return level

    return mutate_level


def make_reset_fn_sample_kinetix_level(
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    ued_params: UEDParams = None,
    physics_engine: PhysicsEngine = None,
):

    ued_params = ued_params or UEDParams()
    physics_engine = physics_engine or PhysicsEngine(static_env_params)

    def reset(rng):
        sampled_level = sample_kinetix_level(rng, physics_engine, env_params, static_env_params, ued_params)

        return sampled_level

    return reset


def make_reset_fn_sample_empty_level(env_params: EnvParams, static_env_params: StaticEnvParams):
    def reset(rng):
        return create_empty_env(static_env_params)

    return reset


def make_vmapped_filtered_level_sampler(
    level_sampler, env_params: EnvParams, static_env_params: StaticEnvParams, config, env, ued_params: UEDParams = None
):
    ued_params = ued_params or UEDParams()

    @partial(jax.jit, static_argnums=(1,))
    def reset(rng, n_samples):
        inner = create_vmapped_filtered_distribution(
            rng,
            level_sampler,
            env_params,
            static_env_params,
            ued_params,
            n_samples,
            env,
            config["filter_levels"],
            config["level_filter_sample_ratio"],
            config["env_size_name"],
            config["level_filter_n_steps"],
        )
        return inner

    return reset


def make_reset_fn_list_of_levels(levels, static_env_params):
    assert len(levels) > 0, "Need to provide at least one level to train on"
    levels_to_reset_to, _ = load_evaluation_levels(levels, static_env_params_override=static_env_params)

    def reset(rng):
        rng, _rng = jax.random.split(rng)
        level_idx = jax.random.randint(_rng, (), 0, len(levels))
        sampled_level = jax.tree.map(lambda x: x[level_idx], levels_to_reset_to)

        return sampled_level

    return reset


ALL_MUTATION_FNS = [
    mutate_add_shape,
    mutate_add_connected_shape,
    mutate_remove_joint,
    mutate_swap_role,
    mutate_toggle_fixture,
    mutate_add_thruster,
    mutate_remove_thruster,
    mutate_remove_shape,
    mutate_change_gravity,
]


def make_reset_fn_from_config(
    config,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    physics_engine: PhysicsEngine = None,
    ued_params: UEDParams = None,
):
    if config["train_level_mode"] == "list":
        reset_fn = make_reset_fn_list_of_levels(config["train_levels_list"], static_env_params)
    elif config["train_level_mode"] == "random":
        reset_fn = make_reset_fn_sample_kinetix_level(env_params, static_env_params, ued_params, physics_engine)
    elif config["train_level_mode"] == "dummy":
        reset_fn = make_reset_fn_sample_empty_level(env_params, static_env_params)
    else:
        raise ValueError("Invalid Reset Function Provided")

    return reset_fn


def test_ued():

    env_params = EnvParams()
    static_env_params = StaticEnvParams()
    ued_params = UEDParams()
    rng = jax.random.PRNGKey(0)
    rng, _rng = jax.random.split(rng)
    state = create_empty_env(env_params, static_env_params)
    state = mutate_add_shape(_rng, state, env_params, static_env_params, ued_params)
    state = mutate_add_connected_shape(_rng, state, env_params, static_env_params, ued_params)
    state = mutate_remove_shape(_rng, state, env_params, static_env_params, ued_params)
    state = mutate_remove_joint(_rng, state, env_params, static_env_params, ued_params)
    state = mutate_swap_role(_rng, state, env_params, static_env_params, ued_params)
    mutate_toggle_fixture(_rng, state, env_params, static_env_params, ued_params)

    print("Successfully did this")


if __name__ == "__main__":
    test_ued()



from functools import partial

import jax
import jax.numpy as jnp
from flax.serialization import to_state_dict
from jax2d.engine import PhysicsEngine

from kinetix.environment.env_state import EnvParams, StaticEnvParams
from kinetix.environment.ued.mutators import (
    mutate_add_connected_shape,
    mutate_add_connected_shape_proper,
    mutate_add_shape,
    mutate_add_thruster,
)
from kinetix.environment.ued.ued_state import UEDParams
from kinetix.environment.utils import create_empty_env, permute_state


# @partial(jax.jit, static_argnums=(1, 3, 5, 6, 7, 8, 9, 10))
def create_vmapped_filtered_distribution(
    rng,
    level_sampler,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    ued_params: UEDParams,
    n_samples: int,
    env,
    do_filter_levels: bool,
    level_filter_sample_ratio: int,
    env_size_name: str,
    level_filter_n_steps: int,
):

    if do_filter_levels and level_filter_n_steps > 0:
        sample_ratio = level_filter_sample_ratio
        n_unfiltered_samples = sample_ratio * n_samples
        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, n_unfiltered_samples)

        # unfiltered_levels = jax.vmap(level_sampler, in_axes=(0, None, None, None, None))(
        #     _rngs, env_params, static_env_params, ued_params, env_size_name
        # )
        unfiltered_levels = jax.vmap(level_sampler, in_axes=(0,))(_rngs)
        #

        # No-op filtering

        def _noop_step(carry, rng):
            states, start_state = carry
            rng, _rng = jax.random.split(rng)
            _rngs = jax.random.split(_rng, n_unfiltered_samples)

            action = jnp.zeros((n_unfiltered_samples, *env.action_space(env_params).shape), dtype=jnp.int32)

            obs, states, reward, done, info = jax.vmap(env.step, in_axes=(0, 0, 0, None, 0))(
                _rngs, states, action, env_params, start_state
            )

            return (states, start_state), (done, reward)

        # Wrap levels
        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, n_unfiltered_samples)
        obsv, unfiltered_levels_wrapped = jax.vmap(env.reset, in_axes=(0, None, 0))(
            _rngs, env_params, unfiltered_levels
        )

        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, level_filter_n_steps)
        _, (done, rewards) = jax.lax.scan(_noop_step, (unfiltered_levels_wrapped, unfiltered_levels_wrapped), xs=_rngs)

        done_indexes = jnp.argmax(done, axis=0)
        done_rewards = rewards[done_indexes, jnp.arange(n_unfiltered_samples)]

        noop_solved_indexes = done_rewards > 0.5
        p = noop_solved_indexes * 0.001 + (1 - noop_solved_indexes) * 1.0
        p /= p.sum()

        rng, _rng = jax.random.split(rng)
        level_indexes = jax.random.choice(
            _rng, jnp.arange(n_unfiltered_samples), shape=(n_samples,), replace=False, p=p
        )

        levels = jax.tree.map(lambda x: x[level_indexes], unfiltered_levels)

    else:
        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, n_samples)

        levels = jax.vmap(level_sampler, in_axes=(0,))(_rngs)

    return levels


@partial(jax.jit, static_argnums=(1, 3, 4))
def sample_kinetix_level(
    rng,
    engine: PhysicsEngine,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    ued_params: UEDParams,
):
    rng, _rng = jax.random.split(rng)
    _rngs = jax.random.split(_rng, 12)

    small_force_no_fixate = static_env_params.num_polygons + static_env_params.num_circles <= 7

    add_shape_n_proposals = ued_params.add_shape_n_proposals
    if static_env_params.num_polygons + static_env_params.num_circles <= 7:
        add_shape_n_proposals = 1  # otherwise we get a very weird XLA bug.

    # Start with empty state
    state = create_empty_env(static_env_params)

    # Set the floor
    prob_of_floor_colour = jnp.array(
        [
            ued_params.floor_prob_normal,
            ued_params.floor_prob_green,
            ued_params.floor_prob_blue,
            ued_params.floor_prob_red,
        ]
    )
    floor_colour = jax.random.choice(_rngs[0], jnp.arange(4), p=prob_of_floor_colour)
    state = state.replace(polygon_shape_roles=state.polygon_shape_roles.at[0].set(floor_colour))

    # When we add shapes we don't want them to collide with already existing shapes
    def _choose_proposal_with_least_collisions(proposals, bias=None):
        rr, cr, cc = jax.vmap(engine.calculate_collision_manifolds)(proposals)

        rr_collisions = jnp.sum(jnp.sum(rr.active.astype(jnp.int32), axis=-1), axis=-1)
        cr_collisions = jnp.sum(cr.active.astype(jnp.int32), axis=-1)
        cc_collisions = jnp.sum(cc.active.astype(jnp.int32), axis=-1)

        all_collisions = jnp.concatenate(
            [rr_collisions[:, None], cr_collisions[:, None], cc_collisions[:, None]], axis=1
        )
        num_collisions = jnp.sum(all_collisions, axis=-1)
        if bias is not None:
            num_collisions = num_collisions + bias

        chosen_addition_idx = jnp.argmin(num_collisions)

        return jax.tree.map(lambda x: x[chosen_addition_idx], proposals)

    def _add_filtered_shape(rng, state, force_no_fixate=False):
        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, add_shape_n_proposals)
        proposed_additions = jax.vmap(mutate_add_shape, in_axes=(0, None, None, None, None, None))(
            _rngs,
            state,
            env_params,
            static_env_params,
            ued_params,
            jnp.logical_or(force_no_fixate, small_force_no_fixate),
        )

        return _choose_proposal_with_least_collisions(proposed_additions)

    def _add_filtered_connected_shape(rng, state, force_rjoint=False):
        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, add_shape_n_proposals)

        proposed_additions, valid = jax.vmap(mutate_add_connected_shape, in_axes=(0, None, None, None, None, None))(
            _rngs, state, env_params, static_env_params, ued_params, force_rjoint
        )

        bias = (jnp.ones(add_shape_n_proposals) - 1 * valid) * ued_params.connect_no_visibility_bias

        return _choose_proposal_with_least_collisions(proposed_additions, bias=bias)

    # Add green and blue - make sure they're not both fixated
    force_green_no_fixate = (jax.random.uniform(_rngs[1]) < 0.5) | (state.polygon_shape_roles[0] == 2)
    state = _add_filtered_shape(_rngs[2], state, force_green_no_fixate)
    state = _add_filtered_shape(_rngs[3], state, ~force_green_no_fixate)

    # Forced controls
    forced_control = jnp.array([[0, 1], [1, 0], [1, 1]])[jax.random.randint(_rngs[4], (), 0, 3)]
    force_thruster, force_motor = forced_control[0], forced_control[1]

    # Forced motor
    state = jax.lax.cond(
        force_motor,
        lambda: _add_filtered_connected_shape(_rngs[5], state, force_rjoint=True),  # force the rjoint
        lambda: _add_filtered_shape(_rngs[6], state),
    )

    # Forced thruster
    state = jax.lax.cond(
        force_thruster,
        lambda: mutate_add_thruster(_rngs[7], state, env_params, static_env_params, ued_params),
        lambda: state,
    )

    # Add rest of shapes
    n_shapes_to_add = (
        static_env_params.num_polygons + static_env_params.num_circles - 3 - static_env_params.num_static_fixated_polys
    )

    def _add_shape(state, rng):
        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, 3)
        shape_add_type = jax.random.choice(
            _rngs[0],
            jnp.arange(3),
            p=jnp.array(
                [ued_params.add_connected_shape_chance, ued_params.add_shape_chance, ued_params.add_no_shape_chance]
            ),
        )

        state = jax.lax.switch(
            shape_add_type,
            [
                lambda: _add_filtered_connected_shape(_rngs[1], state),
                lambda: _add_filtered_shape(_rngs[2], state),
                lambda: state,
            ],
        )

        return state, None

    state, _ = jax.lax.scan(_add_shape, state, jax.random.split(_rngs[8], n_shapes_to_add))

    # Add thrusters
    n_thrusters_to_add = static_env_params.num_thrusters - 1

    def _add_thruster(state, rng):
        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, 3)
        state = jax.lax.cond(
            jax.random.uniform(_rngs[0]) < ued_params.add_thruster_chance,
            lambda: mutate_add_thruster(_rngs[1], state, env_params, static_env_params, ued_params),
            lambda: state,
        )

        return state, None

    state, _ = jax.lax.scan(_add_thruster, state, jax.random.split(_rngs[9], n_thrusters_to_add))

    # Randomly swap green and blue to remove left-right bias
    def _swap_roles(do_swap_roles, roles):
        role1 = roles == 1
        role2 = roles == 2

        swapped_roles = roles * ~(role1 | role2) + role1.astype(int) * 2 + role2.astype(int) * 1
        return jax.lax.select(do_swap_roles, swapped_roles, roles)

    do_swap_roles = jax.random.uniform(_rngs[10], shape=()) < 0.5
    # Don't want to swap if floor is non-standard
    do_swap_roles &= state.polygon_shape_roles[0] == 0
    state = state.replace(
        polygon_shape_roles=_swap_roles(do_swap_roles, state.polygon_shape_roles),
        circle_shape_roles=_swap_roles(do_swap_roles, state.circle_shape_roles),
    )

    return permute_state(_rngs[11], state, static_env_params)


@partial(jax.jit, static_argnums=(2, 4, 5))
def create_random_starting_distribution(
    rng,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    ued_params: UEDParams,
    env_size_name: str,
    controllable=True,
):
    rng, _rng = jax.random.split(rng)
    _rngs = jax.random.split(_rng, 15)
    d = to_state_dict(ued_params)
    ued_params = UEDParams(
        **(
            d
            | dict(
                goal_body_size_factor=2.0,
                thruster_power_multiplier=2.0,
                max_shape_size=0.5,
            )
        ),
    )

    prob_of_large_shapes = 0.05

    ued_params_large_shapes = ued_params.replace(
        max_shape_size=static_env_params.max_shape_size * 1.0, goal_body_size_factor=1.0
    )

    state = create_empty_env(env_params, static_env_params)

    def _get_ued_params(rng):
        rng, _rng, _rng2 = jax.random.split(rng, 3)
        large_shapes = jax.random.uniform(_rng) < prob_of_large_shapes
        params_to_use = jax.tree.map(
            lambda x, y: jax.lax.select(large_shapes, x, y), ued_params_large_shapes, ued_params
        )
        return params_to_use

    def _my_add_shape(rng, state):
        rng, _rng, _rng2 = jax.random.split(rng, 3)
        return mutate_add_shape(_rng, state, env_params, static_env_params, _get_ued_params(_rng2))

    def _my_add_connected_shape(rng, state, **kwargs):
        rng, _rng, _rng2 = jax.random.split(rng, 3)
        return mutate_add_connected_shape_proper(
            _rng, state, env_params, static_env_params, _get_ued_params(_rng2), **kwargs
        )

    # Add the green thing and blue thing
    state = _my_add_shape(_rngs[0], state)
    state = _my_add_shape(_rngs[1], state)
    if controllable:
        # Forced controls
        forced_control = jnp.array([[0, 1], [1, 0], [1, 1]])[jax.random.randint(_rngs[2], (), 0, 3)]
        force_thruster, force_motor = forced_control[0], forced_control[1]

        # Forced motor
        state = jax.lax.cond(
            force_motor,
            lambda: _my_add_connected_shape(_rngs[3], state, force_rjoint=True),  # force the rjoint
            lambda: state,
        )

        # Forced thruster
        state = jax.lax.cond(
            force_thruster,
            lambda: mutate_add_thruster(_rngs[4], state, env_params, static_env_params, ued_params),
            lambda: state,
        )
    return permute_state(_rngs[7], state, static_env_params)



import bz2
import json
import os
import pickle
import re
from typing import Any, Dict, Union

import flax
import flax.serialization
import jax
import jax.numpy as jnp
from flax.traverse_util import unflatten_dict
from jax2d.engine import (
    calculate_collision_matrix,
    get_empty_collision_manifolds,
    get_pairwise_interaction_indices,
    recalculate_mass_and_inertia,
)
from jax2d.sim_state import SimState
from safetensors.flax import load_file

import wandb
from kinetix.environment.env_state import EnvParams, EnvState, StaticEnvParams
from kinetix.environment.utils import create_empty_env, permute_state

BASE_DIR = os.path.realpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "levels"))


def get_correct_path_of_json_level(l: str):
    if not l.endswith(".json"):
        l = l + ".json"
    if os.path.isabs(l) or l.startswith("."):
        return l
    return os.path.join(BASE_DIR, l)


def load_evaluation_levels(eval_levels: list[str], static_env_params_override=None) -> tuple[EnvState, StaticEnvParams]:
    should_permute = [".permute" in l for l in eval_levels]
    eval_levels = [re.sub(r"\.permute\d+", "", l) for l in eval_levels]

    all_levels = []
    all_static_env_params = []
    for l in eval_levels:
        env_state, static_env_params, _ = load_from_json_file(l)
        all_levels.append(env_state)
        all_static_env_params.append(static_env_params)

    if static_env_params_override is not None:
        biggest_static_env_params = static_env_params_override
    else:
        # get biggest static env params
        biggest_static_env_params = jax.tree.map(lambda *x: max(x), *all_static_env_params)
        print(f"Created static_env_params={static_env_params} when loading evaluation levels.")

    all_levels = [expand_env_state(l, biggest_static_env_params) for l in all_levels]
    _rngs = jax.random.split(jax.random.PRNGKey(0), len(all_levels))
    new_ls = [
        permute_state(_rng, l, static_env_params) if sp else l for sp, l, _rng in zip(should_permute, all_levels, _rngs)
    ]
    return stack_list_of_pytrees(new_ls), biggest_static_env_params


def check_if_mass_and_inertia_are_correct(state: SimState, env_params: EnvParams, static_params):
    new = recalculate_mass_and_inertia(state, static_params, state.polygon_densities, state.circle_densities)

    def _check(a, b, shape, name):
        a = jnp.where(shape.active, a, jnp.zeros_like(a))
        b = jnp.where(shape.active, b, jnp.zeros_like(b))

        if not jnp.allclose(a, b):
            idxs = jnp.arange(len(shape.active))[(a != b) & shape.active]
            new_one = a[idxs]
            old_one = b[idxs]
            raise ValueError(
                f"Error: {name} is not the same after loading. Indexes {idxs} are incorrect. New = {new_one} | Before = {old_one}"
            )

    _check(new.polygon.inverse_mass, state.polygon.inverse_mass, state.polygon, "Polygon inverse mass")
    _check(new.circle.inverse_mass, state.circle.inverse_mass, state.circle, "Circle inverse mass")
    _check(new.polygon.inverse_inertia, state.polygon.inverse_inertia, state.polygon, "Polygon inverse inertia")
    _check(new.circle.inverse_inertia, state.circle.inverse_inertia, state.circle, "Circle inverse inertia")
    return True


def save_pickle(filename, state):
    with open(filename, "wb") as f:
        pickle.dump(state, f)


def expand_env_state(env_state: EnvState, static_env_params: StaticEnvParams, ignore_collision_matrix=False):

    num_rects = len(env_state.polygon.position)
    num_circles = len(env_state.circle.position)
    num_joints = len(env_state.joint.a_index)
    num_thrusters = len(env_state.thruster.object_index)

    def _add_dummy(num_to_add, obj):
        return jax.tree.map(
            lambda current: jnp.concatenate(
                [current, jnp.zeros((num_to_add, *current.shape[1:]), dtype=current.dtype)], axis=0
            ),
            obj,
        )

    does_need_to_change = False
    added_rects = 0

    if (
        num_rects > static_env_params.num_polygons
        or num_circles > static_env_params.num_circles
        or num_joints > static_env_params.num_joints
    ):
        raise Exception(
            f"The current static_env_params is too small to accommodate the loaded env_state (needs num_rects={num_rects}, num_circles={num_circles}, num_joints={num_joints} but current is {static_env_params.num_polygons}, {static_env_params.num_circles}, {static_env_params.num_joints})."
        )

    if num_rects < static_env_params.num_polygons:
        added_rects = static_env_params.num_polygons - num_rects
        does_need_to_change = True
        env_state = env_state.replace(
            polygon=_add_dummy(added_rects, env_state.polygon),
            polygon_shape_roles=_add_dummy(added_rects, env_state.polygon_shape_roles),
            polygon_highlighted=_add_dummy(added_rects, env_state.polygon_highlighted),
            polygon_densities=_add_dummy(added_rects, env_state.polygon_densities),
        )

    if num_circles < static_env_params.num_circles:
        does_need_to_change = True
        n_to_add = static_env_params.num_circles - num_circles
        env_state = env_state.replace(
            circle=_add_dummy(n_to_add, env_state.circle),
            circle_shape_roles=_add_dummy(n_to_add, env_state.circle_shape_roles),
            circle_highlighted=_add_dummy(n_to_add, env_state.circle_highlighted),
            circle_densities=_add_dummy(n_to_add, env_state.circle_densities),
        )

    if num_joints < static_env_params.num_joints:
        does_need_to_change = True
        n_to_add = static_env_params.num_joints - num_joints
        env_state = env_state.replace(
            joint=_add_dummy(n_to_add, env_state.joint),
            motor_bindings=_add_dummy(n_to_add, env_state.motor_bindings),
            motor_auto=_add_dummy(n_to_add, env_state.motor_auto),
        )

    if num_thrusters < static_env_params.num_thrusters:
        does_need_to_change = True
        n_to_add = static_env_params.num_thrusters - num_thrusters
        env_state = env_state.replace(
            thruster=_add_dummy(n_to_add, env_state.thruster),
            thruster_bindings=_add_dummy(n_to_add, env_state.thruster_bindings),
        )

    # This fixes the indices
    def _modify_index(old_indices):
        return jnp.where(old_indices >= num_rects, old_indices + added_rects, old_indices)

    if added_rects > 0:
        env_state = env_state.replace(
            joint=env_state.joint.replace(
                a_index=_modify_index(env_state.joint.a_index),
                b_index=_modify_index(env_state.joint.b_index),
            ),
            thruster=env_state.thruster.replace(
                object_index=_modify_index(env_state.thruster.object_index),
            ),
        )
    # Double check the collision manifolds are fine
    if does_need_to_change or 1:
        # print("Loading but changing the shapes to match the current static params.")
        acc_rr_manifolds, acc_cr_manifolds, acc_cc_manifolds = get_empty_collision_manifolds(static_env_params)
        env_state = env_state.replace(
            collision_matrix=(
                env_state.collision_matrix
                if ignore_collision_matrix
                else calculate_collision_matrix(static_env_params, env_state.joint)
            ),
            acc_rr_manifolds=acc_rr_manifolds,
            acc_cr_manifolds=acc_cr_manifolds,
            acc_cc_manifolds=acc_cc_manifolds,
        )
    return env_state


def load_world_state_pickle(filename, env_params=None, static_env_params=None):
    static_params = static_env_params or StaticEnvParams()
    with open(filename, "rb") as f:
        state: SimState = pickle.load(f)
        state = jax.tree.map(lambda x: jnp.nan_to_num(x), state)
        # Check if the mass and inertia are reasonable.
        check_if_mass_and_inertia_are_correct(state, env_params or EnvParams(), static_params)

    # Now check if the shapes are correct
    return expand_env_state(state, static_params)


def stack_list_of_pytrees(list_of_pytrees):
    v = jax.tree.map(lambda x: jnp.expand_dims(x, 0), list_of_pytrees[0])
    for l in list_of_pytrees[1:]:
        v = jax.tree.map(lambda x, y: jnp.concatenate([x, jnp.expand_dims(y, 0)], axis=0), v, l)
    return v


def get_env_state_from_json(json_filename) -> EnvState:
    env_state, _, _ = load_from_json_file(json_filename)
    return env_state


def my_load_file(filename):
    data = bz2.BZ2File(filename, "rb")
    data = pickle.load(data)
    return data


def my_save_file(obj, filename):
    with bz2.BZ2File(filename, "w") as f:
        pickle.dump(obj, f)


def save_params(params: Dict, filename: Union[str, os.PathLike]) -> None:
    my_save_file(params, filename)


def load_params(filename: Union[str, os.PathLike]) -> Dict:
    return my_load_file(filename)


def load_params_from_wandb_artifact_path(checkpoint_name):
    api = wandb.Api()
    name = api.artifact(checkpoint_name).download()
    network_params = load_params(name + "/model.pbz2")
    return network_params


def save_params_to_wandb(params, timesteps, config):
    if config["checkpoint_human_numbers"]:
        timesteps = str(round(timesteps / 1e9)) + "B"

    run_name = config["run_name"] + "-" + str(config["random_hash"]) + "-" + str(timesteps)
    save_dir = os.path.join(config["save_path"], run_name)
    os.makedirs(save_dir, exist_ok=True)
    save_params(params, f"{save_dir}/model.pbz2")

    # upload this to wandb as an artifact
    artifact = wandb.Artifact(f"{run_name}-checkpoint", type="checkpoint")
    artifact.add_file(f"{save_dir}/model.pbz2")
    artifact.save()
    print(f"Parameters of model saved in {save_dir}/model.pbz2")


def load_params_wandb_artifact_path_full_model(checkpoint_name):
    api = wandb.Api()
    name = api.artifact(checkpoint_name).download()
    all_dict = load_params(name + "/full_model.pbz2")
    return all_dict["params"]


def load_train_state_from_wandb_artifact_path(train_state, checkpoint_name, load_only_params=False):
    api = wandb.Api()
    name = api.artifact(checkpoint_name).download()
    all_dict = load_params(name + "/full_model.pbz2")
    train_state = train_state.replace(params=all_dict["params"])
    if not load_only_params:
        train_state = train_state.replace(
            # step=all_dict["step"],
            opt_state=all_dict["opt_state"]
        )
    return train_state


def save_params_to_wandb(params, timesteps, config):
    return save_dict(params, timesteps, config, "params")


def save_dict(dict, timesteps, config, name, save_to_wandb: bool = True):
    timesteps = str(round(timesteps / 1e9)) + "B"
    run_name = config["run_name"] + "-" + str(config["random_hash"]) + "-" + str(timesteps)
    save_dir = os.path.join(config["save_path"], run_name)
    os.makedirs(save_dir, exist_ok=True)
    save_params(dict, f"{save_dir}/{name}.pbz2")

    # upload this to wandb as an artifact
    if save_to_wandb:
        artifact = wandb.Artifact(f"{run_name}-checkpoint", type="checkpoint")
        artifact.add_file(f"{save_dir}/{name}.pbz2")
        artifact.save()
    print(f"Parameters of model saved in {save_dir}/{name}.pbz2")


def save_model(train_state, timesteps, config, is_final=False, save_to_wandb: bool = True):
    dict_to_use = {"step": train_state.step, "params": train_state.params, "opt_state": train_state.opt_state}
    step = int(train_state.step)
    if config["economical_saving"]:
        if step in [2048, 10240, 40960, 81920] or is_final:
            save_dict(dict_to_use, timesteps, config, "full_model", save_to_wandb=save_to_wandb)
        else:
            print("Not saving model because step is", step)
    else:
        save_dict(dict_to_use, timesteps, config, "full_model", save_to_wandb=save_to_wandb)


def import_env_state_from_json(json_file: dict[str, Any]) -> tuple[EnvState, StaticEnvParams, EnvParams]:
    def normalise(k, v):
        if k == "screen_dim":
            return v
        if type(v) == dict and "0" in v:
            return jnp.array([normalise(k, v[str(i)]) for i in range(len(v))])
        return v

    env_state = json_file["env_state"]
    env_params = json_file["env_params"]
    static_env_params = json_file["static_env_params"]
    env_params_target = EnvParams()
    static_env_params_target = StaticEnvParams()
    new_env_params = flax.serialization.from_state_dict(
        env_params_target, {k: normalise(k, v) for k, v in env_params.items()}
    )
    norm_static = {k: normalise(k, v) for k, v in static_env_params.items()}
    # norm_static["screen_dim"] = tuple(static_env_params_target.screen_dim)
    norm_static["downscale"] = static_env_params_target.downscale
    # print(
    #     static_env_params_target,
    # )
    new_static_env_params = flax.serialization.from_state_dict(static_env_params_target, norm_static)
    new_static_env_params = new_static_env_params.replace(screen_dim=static_env_params_target.screen_dim)

    env_state_target = create_empty_env(new_static_env_params)

    def astype(x, all):
        return jnp.astype(x, all.dtype)

    def _load_rigidbody(env_state_target, i, is_poly):

        to_load_from: dict[str, Any] = env_state["circle" if not is_poly else "polygon"][i]
        role = to_load_from.pop("role")
        density = to_load_from.pop("density")
        if "highlighted" in to_load_from:
            _ = to_load_from.pop("highlighted")
        new_obj = flax.serialization.from_state_dict(
            jax.tree.map(lambda x: x[i], env_state_target.circle if not is_poly else env_state_target.polygon),
            {k: normalise(k, v) for k, v in to_load_from.items()},
        )

        if is_poly:
            env_state_target = env_state_target.replace(
                polygon_shape_roles=env_state_target.polygon_shape_roles.at[i].set(role),
                polygon_densities=env_state_target.polygon_densities.at[i].set(density),
                polygon=jax.tree.map(
                    lambda all, new: all.at[i].set(astype(new, all)), env_state_target.polygon, new_obj
                ),
            )
        else:
            env_state_target = env_state_target.replace(
                circle_shape_roles=env_state_target.circle_shape_roles.at[i].set(role),
                circle_densities=env_state_target.circle_densities.at[i].set(density),
                circle=jax.tree.map(lambda all, new: all.at[i].set(astype(new, all)), env_state_target.circle, new_obj),
            )
        return env_state_target

    # Now load the env state:
    for i in range(new_static_env_params.num_circles):
        env_state_target = _load_rigidbody(env_state_target, i, False)
    for i in range(new_static_env_params.num_polygons):
        env_state_target = _load_rigidbody(env_state_target, i, True)

    for i in range(new_static_env_params.num_joints):
        to_load_from = env_state["joint"][i]
        motor_binding = to_load_from.pop("motor_binding")
        new_obj = flax.serialization.from_state_dict(
            jax.tree.map(lambda x: x[i], env_state_target.joint), {k: normalise(k, v) for k, v in to_load_from.items()}
        )
        env_state_target = env_state_target.replace(
            joint=jax.tree.map(lambda all, new: all.at[i].set(astype(new, all)), env_state_target.joint, new_obj),
            motor_bindings=env_state_target.motor_bindings.at[i].set(motor_binding),
        )

    for i in range(new_static_env_params.num_thrusters):
        to_load_from = env_state["thruster"][i]
        thruster_binding = to_load_from.pop("thruster_binding")
        new_obj = flax.serialization.from_state_dict(
            jax.tree.map(lambda x: x[i], env_state_target.thruster),
            {k: normalise(k, v) for k, v in to_load_from.items()},
        )

        env_state_target = env_state_target.replace(
            thruster=jax.tree.map(lambda all, new: all.at[i].set(astype(new, all)), env_state_target.thruster, new_obj),
            thruster_bindings=env_state_target.thruster_bindings.at[i].set(thruster_binding),
        )

    env_state_target = env_state_target.replace(
        collision_matrix=flax.serialization.from_state_dict(
            env_state_target.collision_matrix, normalise("collision_matrix", env_state["collision_matrix"])
        )
    )

    for i in range(env_state_target.acc_rr_manifolds.active.shape[0]):
        a = flax.serialization.from_state_dict(
            jax.tree.map(lambda x: x[i], env_state_target.acc_rr_manifolds),
            {k: normalise(k, v) for k, v in env_state["acc_rr_manifolds"][i].items()},
        )
        b = flax.serialization.from_state_dict(
            jax.tree.map(lambda x: x[i], env_state_target.acc_rr_manifolds),
            {k: normalise(k, v) for k, v in env_state["acc_rr_manifolds"][i + 1].items()},
        )
        env_state_target = env_state_target.replace(
            acc_rr_manifolds=jax.tree.map(
                lambda all, new: all.at[i].set(astype(new, all)), env_state_target.acc_rr_manifolds, a
            ),
        )
        env_state_target.replace(
            acc_rr_manifolds=jax.tree.map(
                lambda all, new: all.at[i + 1].set(astype(new, all)), env_state_target.acc_rr_manifolds, b
            )
        )
    for i in range(env_state_target.acc_cr_manifolds.active.shape[0]):
        a = flax.serialization.from_state_dict(
            jax.tree.map(lambda x: x[i], env_state_target.acc_cr_manifolds),
            {k: normalise(k, v) for k, v in env_state["acc_cr_manifolds"][i].items()},
        )
        env_state_target = env_state_target.replace(
            acc_cr_manifolds=jax.tree.map(
                lambda all, new: all.at[i].set(astype(new, all)), env_state_target.acc_cr_manifolds, a
            ),
        )
    for i in range(env_state_target.acc_cc_manifolds.active.shape[0]):
        a = flax.serialization.from_state_dict(
            jax.tree.map(lambda x: x[i], env_state_target.acc_cc_manifolds),
            {k: normalise(k, v) for k, v in env_state["acc_cc_manifolds"][i].items()},
        )
        env_state_target = env_state_target.replace(
            acc_cc_manifolds=jax.tree.map(
                lambda all, new: all.at[i].set(astype(new, all)), env_state_target.acc_cc_manifolds, a
            ),
        )

    env_state_target = env_state_target.replace(
        collision_matrix=calculate_collision_matrix(new_static_env_params, env_state_target.joint)
    )

    return (
        env_state_target,
        new_static_env_params,
        new_env_params.replace(max_timesteps=env_params_target.max_timesteps),
    )


def export_env_state_to_json(
    filename: str, env_state: EnvState, static_env_params: StaticEnvParams, env_params: EnvParams
):
    json_to_save = {
        "polygon": [],
        "circle": [],
        "joint": [],
        "thruster": [],
        "collision_matrix": flax.serialization.to_state_dict(env_state.collision_matrix.tolist()),
        "acc_rr_manifolds": [],
        "acc_cr_manifolds": [],
        "acc_cc_manifolds": [],
        "gravity": flax.serialization.to_state_dict(env_state.gravity.tolist()),
    }

    def _rigidbody_to_json(index: int, is_poly):
        main_arr = env_state.polygon if is_poly else env_state.circle
        c = jax.tree.map(lambda x: x[index].tolist(), main_arr)
        roles = env_state.polygon_shape_roles if is_poly else env_state.circle_shape_roles
        densities = env_state.polygon_densities if is_poly else env_state.circle_densities
        highlighted = env_state.polygon_highlighted if is_poly else env_state.circle_highlighted

        d = flax.serialization.to_state_dict(c)
        d["role"] = roles[index].tolist()
        d["density"] = densities[index].tolist()
        d["highlighted"] = highlighted[index].tolist()
        return d

    def _joint_to_json(i):
        joint = jax.tree.map(lambda x: x[i].tolist(), env_state.joint)
        d = flax.serialization.to_state_dict(joint)
        d["motor_binding"] = env_state.motor_bindings[i].tolist()
        return d

    def _thruster_to_json(i):
        thruster = jax.tree.map(lambda x: x[i].tolist(), env_state.thruster)
        d = flax.serialization.to_state_dict(thruster)
        d["thruster_binding"] = env_state.thruster_bindings[i].tolist()
        return d

    for i in range(static_env_params.num_circles):
        json_to_save["circle"].append(_rigidbody_to_json(i, False))
    for i in range(static_env_params.num_polygons):
        json_to_save["polygon"].append(_rigidbody_to_json(i, True))
    for i in range(static_env_params.num_joints):
        json_to_save["joint"].append(_joint_to_json(i))
    for i in range(static_env_params.num_thrusters):
        json_to_save["thruster"].append(_thruster_to_json(i))

    ncc, ncr, nrr, circle_circle_pairs, circle_rect_pairs, rect_rect_pairs = get_pairwise_interaction_indices(
        static_env_params
    )
    for i in range(nrr):
        a = jax.tree.map(lambda x: x[i, 0].tolist(), env_state.acc_rr_manifolds)
        b = jax.tree.map(lambda x: x[i, 1].tolist(), env_state.acc_rr_manifolds)
        json_to_save["acc_rr_manifolds"].append(flax.serialization.to_state_dict(a))
        json_to_save["acc_rr_manifolds"].append(flax.serialization.to_state_dict(b))
    for i in range(ncr):
        a = jax.tree.map(lambda x: x[i].tolist(), env_state.acc_cr_manifolds)
        json_to_save["acc_cr_manifolds"].append(flax.serialization.to_state_dict(a))

    for i in range(ncc):
        a = jax.tree.map(lambda x: x[i].tolist(), env_state.acc_cc_manifolds)
        json_to_save["acc_cc_manifolds"].append(flax.serialization.to_state_dict(a))

    to_save = {
        "env_state": json_to_save,
        "env_params": flax.serialization.to_state_dict(
            jax.tree.map(lambda x: x.tolist() if type(x) == jnp.ndarray else x, env_params)
        ),
        "static_env_params": flax.serialization.to_state_dict(
            jax.tree.map(lambda x: x.tolist() if type(x) == jnp.ndarray else x, static_env_params)
        ),
    }
    with open(filename, "w+") as f:
        json.dump(to_save, f)

    return to_save


def load_from_json_file(filename):
    filename = get_correct_path_of_json_level(filename)
    assert os.path.exists(
        filename
    ), f"File {filename} does not exist. Please ensure the path exists from where you are running your script. Alternatively, please provide an absolute path"
    with open(filename, "r") as f:
        return import_env_state_from_json(json.load(f))



from kinetix.environment.spaces import ActionType, ObservationType
from kinetix.models.actor_critic import (
    ActorCriticPermutationInvariantSymbolicRNN,
    ActorCriticPixelsRNN,
    ActorCriticSymbolicRNN,
)
from kinetix.models.transformer_model import ActorCriticTransformer


def make_network_from_config(env, env_params, config, network_kws={}):

    action_mode = config["action_type"]
    action_dim = (
        env.action_space(env_params).shape[0]
        if action_mode == ActionType.CONTINUOUS
        else env.action_space(env_params).n
    )
    if "hybrid_action_continuous_dim" not in network_kws:
        network_kws["hybrid_action_continuous_dim"] = action_dim

    if "multi_discrete_number_of_dims_per_distribution" not in network_kws:
        num_joint_bindings = config["static_env_params"]["num_motor_bindings"]
        num_thruster_bindings = config["static_env_params"]["num_thruster_bindings"]
        network_kws["multi_discrete_number_of_dims_per_distribution"] = [3 for _ in range(num_joint_bindings)] + [
            2 for _ in range(num_thruster_bindings)
        ]
    network_kws["recurrent"] = config.get("recurrent_model", True)

    obs_type = config["observation_type"]
    assert obs_type in ObservationType and type(obs_type) != str
    if obs_type == ObservationType.PIXELS:
        cls_to_use = ActorCriticPixelsRNN
    elif obs_type == ObservationType.SYMBOLIC_FLAT:
        cls_to_use = ActorCriticSymbolicRNN

    if obs_type == ObservationType.SYMBOLIC_FLAT_PADDED or (
        obs_type == ObservationType.SYMBOLIC_ENTITY and config["permutation_invariant_mlp"]
    ):
        network = ActorCriticPermutationInvariantSymbolicRNN(
            action_dim=action_dim,
            fc_layer_width=config["fc_layer_width"],
            fc_layer_depth=config["fc_layer_depth"],
            action_mode=action_mode,
            activation=config["activation"],
            permutation_invariant=True,
            num_heads=config["num_heads"],
            symbolic_embedding_dim=config["symbolic_embedding_dim"],
            preprocess_separately=config["permutation_invariant_mlp"],
            encoder_size=config["encoder_size"],
            **network_kws,
        )
    elif obs_type == ObservationType.SYMBOLIC_ENTITY:
        network = ActorCriticTransformer(
            action_dim=action_dim,
            fc_layer_width=config["fc_layer_width"],
            fc_layer_depth=config["fc_layer_depth"],
            action_mode=action_mode,
            num_heads=config["num_heads"],
            transformer_depth=config["transformer_depth"],
            transformer_size=config["transformer_size"],
            transformer_encoder_size=config["transformer_encoder_size"],
            aggregate_mode=config["aggregate_mode"],
            full_attention_mask=config["full_attention_mask"],
            activation=config["activation"],
            **network_kws,
        )
    else:
        network = cls_to_use(
            action_dim,
            fc_layer_width=config["fc_layer_width"],
            fc_layer_depth=config["fc_layer_depth"],
            activation=config["activation"],
            action_mode=action_mode,
            **network_kws,
        )

    return network



from typing import Callable
import jax
import jax.numpy as jnp
from flax import struct
from jax2d.engine import get_pairwise_interaction_indices

from kinetix.environment.env_state import EnvParams, EnvState, StaticEnvParams
from kinetix.environment.utils import create_empty_env
from kinetix.render import inverse_renderer_common
from kinetix.render.renderer_symbolic_common import (
    make_circle_features,
    make_joint_features,
    make_polygon_features,
    make_thruster_features,
)
from kinetix.render.renderer_symbolic_flat import _get_symbolic_features
from jax2d.engine import calculate_collision_matrix


@struct.dataclass
class EntityObservation:
    circles: jnp.ndarray
    polygons: jnp.ndarray
    joints: jnp.ndarray
    thrusters: jnp.ndarray

    circle_mask: jnp.ndarray
    polygon_mask: jnp.ndarray
    joint_mask: jnp.ndarray
    thruster_mask: jnp.ndarray
    attention_mask: jnp.ndarray
    # collision_mask: jnp.ndarray

    joint_indexes: jnp.ndarray
    thruster_indexes: jnp.ndarray


def make_render_entities(env_params, static_params, ignore_attention_mask=False):
    _, _, _, circle_circle_pairs, circle_rect_pairs, rect_rect_pairs = get_pairwise_interaction_indices(static_params)
    circle_rect_pairs = circle_rect_pairs.at[:, 0].add(static_params.num_polygons)
    circle_circle_pairs = circle_circle_pairs + static_params.num_polygons

    def render_entities(state: EnvState):
        state = jax.tree.map(lambda x: jnp.nan_to_num(x), state)

        joint_features, joint_indexes, joint_mask = make_joint_features(state, env_params, static_params)
        thruster_features, thruster_indexes, thruster_mask = make_thruster_features(state, env_params, static_params)

        poly_nodes, poly_mask = make_polygon_features(state, env_params, static_params)
        circle_nodes, circle_mask = make_circle_features(state, env_params, static_params)

        def _add_grav(nodes):
            return jnp.concatenate(
                [nodes, jnp.zeros((nodes.shape[0], 1)) + state.gravity[1] / 10], axis=-1
            )  # add gravity to each shape's embedding

        poly_nodes = _add_grav(poly_nodes)
        circle_nodes = _add_grav(circle_nodes)

        # Shape of something like (NPoly + NCircle + 2 * NJoint + NThruster )
        mask_flat_shapes = jnp.concatenate([poly_mask, circle_mask], axis=0)
        num_shapes = static_params.num_polygons + static_params.num_circles

        def make_n_squared_mask(val):
            # val has shape N of bools.
            N = val.shape[0]
            A = jnp.eye(N, N, dtype=bool)  # also have things attend to themselves
            # Make the shapes fully connected
            full_mask = A.at[:num_shapes, :num_shapes].set(jnp.ones((num_shapes, num_shapes), dtype=bool))

            one_hop_connected = jnp.zeros((N, N), dtype=bool)
            one_hop_connected = one_hop_connected.at[joint_indexes[:, 0], joint_indexes[:, 1]].set(True)
            one_hop_connected = one_hop_connected.at[0, 0].set(False)  # invalid joints have indices of (0, 0)

            multi_hop_connected = jnp.logical_not(state.collision_matrix)

            collision_mask = state.collision_matrix

            # where val is false, we want to mask out the row and column.
            full_mask = full_mask & (val[:, None]) & (val[None, :])
            collision_mask = collision_mask & (val[:, None]) & (val[None, :])
            multi_hop_connected = multi_hop_connected & (val[:, None]) & (val[None, :])
            one_hop_connected = one_hop_connected & (val[:, None]) & (val[None, :])
            collision_manifold_mask = jnp.zeros_like(collision_mask)

            def _set(collision_manifold_mask, pairs, active):
                return collision_manifold_mask.at[
                    pairs[:, 0],
                    pairs[:, 1],
                ].set(active)

            collision_manifold_mask = _set(
                collision_manifold_mask,
                rect_rect_pairs,
                jnp.logical_or(state.acc_rr_manifolds.active[..., 0], state.acc_rr_manifolds.active[..., 1]),
            )

            collision_manifold_mask = _set(collision_manifold_mask, circle_rect_pairs, state.acc_cr_manifolds.active)
            collision_manifold_mask = _set(collision_manifold_mask, circle_circle_pairs, state.acc_cc_manifolds.active)
            collision_manifold_mask = collision_manifold_mask & (val[:, None]) & (val[None, :])

            return jnp.concatenate(
                [full_mask[None], multi_hop_connected[None], one_hop_connected[None], collision_manifold_mask[None]],
                axis=0,
            )

        if ignore_attention_mask:
            mask_n_squared = None
        else:
            mask_n_squared = make_n_squared_mask(mask_flat_shapes)

        return EntityObservation(
            circles=circle_nodes,
            polygons=poly_nodes,
            joints=joint_features,
            thrusters=thruster_features,
            circle_mask=circle_mask,
            polygon_mask=poly_mask,
            joint_mask=joint_mask,
            thruster_mask=thruster_mask,
            attention_mask=mask_n_squared,
            joint_indexes=joint_indexes,
            thruster_indexes=thruster_indexes,
        )

    return render_entities


def make_inverse_render_entity(
    env_state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams
) -> Callable[[jnp.ndarray], EnvState]:
    """This creates an inverse renderer, which takes in an observation and returns an EnvState.

    Args:
        env_state (EnvState): This is a dummy env state, and is used to get the shapes of the various arrays.

    Returns:
        Callable[[jnp.ndarray], EnvState]: Maps symbolic entity observation to Env
    """

    (
        nshapes,
        *_,
    ) = _get_symbolic_features(env_state, env_params, static_env_params)

    def symbolic_entity_obs_to_env_state(obs: EntityObservation) -> EnvState:

        env_state = create_empty_env(static_env_params)
        polygons, polygon_shape_roles, polygon_densities = inverse_renderer_common.features_to_rigidbody(
            env_params, static_env_params, obs.polygons, static_env_params.num_polygons, is_circle=False
        )

        polygons = polygons.replace(
            active=obs.polygon_mask,
            collision_mode=polygons.collision_mode.at[: static_env_params.num_static_fixated_polys].set(2),
        )

        circles, circle_shape_roles, circle_densities = inverse_renderer_common.features_to_rigidbody(
            env_params, static_env_params, obs.circles, static_env_params.num_circles, is_circle=True
        )
        circles = circles.replace(active=obs.circle_mask)

        env_state = env_state.replace(
            polygon=polygons,
            circle=circles,
            polygon_shape_roles=polygon_shape_roles,
            circle_shape_roles=circle_shape_roles,
            polygon_densities=polygon_densities,
            circle_densities=circle_densities,
        )

        joints, motor_bindings = inverse_renderer_common.features_to_joint(
            static_env_params,
            env_state,
            obs.joints[: static_env_params.num_joints],
            nshapes,
            joint_idxs_override=obs.joint_indexes[: static_env_params.num_joints],
        )

        joints = joints.replace(
            active=obs.joint_mask[: static_env_params.num_joints],
        )

        thrusters, thruster_bindings = inverse_renderer_common.features_to_thruster(
            static_env_params, env_state, obs.thrusters, nshapes, thruster_idxs_override=obs.thruster_indexes
        )
        thrusters = thrusters.replace(active=obs.thruster_mask, object_index=obs.thruster_indexes)

        env_state = env_state.replace(
            joint=joints, thruster=thrusters, motor_bindings=motor_bindings, thruster_bindings=thruster_bindings
        )

        return env_state.replace(
            collision_matrix=calculate_collision_matrix(static_env_params, env_state.joint),
        )

    return symbolic_entity_obs_to_env_state
